# %%
"""
Tutorial 1: Results
===================

In this tutorial, we'll cover all of the output that comes from a `Phase` in the form of a `Result` object.

we'll use the same problem of fitting 1D profiles to noisy data, with the source code we use to do this in chapter 2
an adaption of the source code we completed chapter 1 using. It has new functionality which we'll cover throughout this
chapter, but for this tutorial the source code is unchanged.

We used this object at various points in the previous chapter, and the bulk of material covered here is described in
the example script `autofit_workspace/examples/simple/result.py`. Nevertheless, it is a good idea to refresh ourselves
about how results in **PyAutoFit** work before covering more advanced material.
"""


# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_2_results import src as htf

import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_2",
)

# %%
"""
Now, lets create a `Dataset` and fit it using a `Phase`, in an identical fashion to the previous chapter.
"""

# %%
from howtofit.simulators.chapter_2 import gaussian_x1

# %%
"""
When we fit the `Dataset``., we omit the data-trimming demonstrated in the previous tutorial.
"""

# %%
settings_masked_dataset = htf.SettingsMaskedDataset(
    data_trim_left=None, data_trim_right=None
)

settings = htf.SettingsPhase(settings_masked_dataset=settings_masked_dataset)

# %%
"""The code below creates the `Dataset` and `mask` as per usual."""

# %%
data = gaussian_x1.data
noise_map = gaussian_x1.noise_map
dataset = htf.Dataset(data=data, noise_map=noise_map)
mask = np.full(fill_value=False, shape=dataset.data.shape)

print(
    f"Emcee has begun running - checkout the "
    f"autofit_workspace/howtofit/chapter_2_results/output/phase_t1 folder for live "
    f"output of the results. This Jupyter notebook cell with progress once Emcee has completed - this could take a "
    f"few minutes!"
)

phase = htf.Phase(
    phase_name="phase_t1",
    profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
    settings=settings,
    search=af.DynestyStatic(),
)

"""Note that we pass the info to the phase when we run it, so that the aggregator can make it accessible."""
result = phase.run(dataset=dataset, mask=mask)


# %%
"""
__Result__

Here, we'll look in detail at what information is contained in the `Result`.

It contains a `Samples` object, which contains information on the `NonLinearSearch`, for example the parameters. 

The parameters are stored as a a list of lists, where:

 - The outer list is the size of the total number of samples.
 - The inner list is the size of the number of free parameters in the fit.

"""

# %%
samples = result.samples
print("All Parameters:")
print(samples.parameters)
print("Sample 10`s third parameter value (Gaussian -> sigma)")
print(samples.parameters[9][1], "\n")

# %%
"""
The `Samples` class also contains the `log_likelihoods`, `log_priors`, `log_posteriors` and `weights` of every 
accepted sample, where:

 - A `log_likelihood` is the value evaluated from the `log_likelihood_function` (e.g. -0.5 * `chi_squared` + the 
 `noise_normalization`).

 - The `log_prior` encodes information on how the priors on the parameters maps the `log_likelihood` value to the 
  `log_posterior` value.

 - The `log_posterior` is `log_likelihood` + `log_prior`.

 - The `weights` gives information on how samples should be combined to estimate the posterior. The values 
   depend on the `NonLinearSearch` used, for MCMC samples they are all 1 (e.g. all weighted equally).
"""

# %%
print("All Log Likelihoods:")
print(samples.log_likelihoods)
print("All Log Priors:")
print(samples.log_priors)
print("All Log Posteriors:")
print(samples.log_posteriors)
print("All Sample Weights:")
print(samples.weights, "\n")

# %%
"""
The `Samples` contain many useful vectors, including the the maximum log likelihood and posterior values:
"""

# %%
max_log_likelihood_vector = samples.max_log_likelihood_vector
max_log_posterior_vector = samples.max_log_posterior_vector

print("Maximum Log Likelihood Vector:")
print(max_log_likelihood_vector)
print("Maximum Log Posterior Vector:")
print(max_log_posterior_vector, "\n")

# %%
"""
This provides us with lists of all model parameters. However, this isn't that much use - which values correspond to 
which parameters?

The list of parameter names are available as a property of the `Samples`, as are labels which can be used for 
labeling figures.
"""

# %%
print(samples.model_component_and_parameter_names)
print(samples.parameter_labels)
print("\n")

# %%
"""
It is more useful to return the `Result`'s as an `instance`, which is an instance of the `model` using the Python 
classes used to compose it.
"""

# %%
max_log_likelihood_instance = samples.max_log_likelihood_instance

# %%
"""
A `model instance` contains all the model components of our fit, which for the fit above was a single `Gaussian`
profile (the word `gaussian` comes from what we called it in the _CollectionPriorModel_ above).
"""

# %%
print(max_log_likelihood_instance.profiles.gaussian)

# %%
"""
We can unpack the parameters of the `Gaussian` to reveal the `max_log_likelihood_instance`:
"""

# %%
print("Max Log Likelihood `Gaussian` Instance:")
print("Centre = ", max_log_likelihood_instance.profiles.gaussian.centre)
print("Intensity = ", max_log_likelihood_instance.profiles.gaussian.intensity)
print("Sigma = ", max_log_likelihood_instance.profiles.gaussian.sigma, "\n")

# %%
"""
For our example problem of fitting a 1D `Gaussian` profile, this makes it straight forward to plot the 
`max_log_likelihood_instance`:
"""

# %%
model_data = samples.max_log_likelihood_instance.profiles.gaussian.profile_from_xvalues(
    xvalues=np.arange(data.shape[0])
)

import matplotlib.pyplot as plt

plt.plot(range(data.shape[0]), data)
plt.plot(range(data.shape[0]), model_data)
plt.title("Illustrative model fit to 1D `Gaussian` profile data.")
plt.xlabel("x values of profile")
plt.ylabel("Profile intensity")
plt.show()
plt.close()


# %%
"""
We can also access the `median PDF` model, which is the model computed by marginalizing over the samples of every 
parameter in 1D and taking the median of this PDF.

The `median_pdf_vector` is readily available from the `Samples` object for you convenience.
"""

# %%
median_pdf_vector = samples.median_pdf_vector
print("Median PDF Vector:")
print(median_pdf_vector, "\n")

median_pdf_instance = samples.median_pdf_instance
print("Median PDF `Gaussian` Instance:")
print("Centre = ", median_pdf_instance.profiles.gaussian.centre)
print("Intensity = ", median_pdf_instance.profiles.gaussian.intensity)
print("Sigma = ", median_pdf_instance.profiles.gaussian.sigma, "\n")


# %%
"""
The `Samples` class also provides methods for computing the error estimates of all parameters at an input sigma 
confidence limit, which can be returned as the values of the parameters including their errors or the size of the 
errors on each parameter:
"""

# %%
vector_at_upper_sigma = samples.vector_at_upper_sigma(sigma=3.0)
vector_at_lower_sigma = samples.vector_at_lower_sigma(sigma=3.0)

print("Upper Parameter values w/ error (at 3.0 sigma confidence):")
print(vector_at_upper_sigma)
print("lower Parameter values w/ errors (at 3.0 sigma confidence):")
print(vector_at_lower_sigma, "\n")

error_vector_at_upper_sigma = samples.error_vector_at_upper_sigma(sigma=3.0)
error_vector_at_lower_sigma = samples.error_vector_at_lower_sigma(sigma=3.0)

print("Upper Error values (at 3.0 sigma confidence):")
print(error_vector_at_upper_sigma)
print("lower Error values (at 3.0 sigma confidence):")
print(error_vector_at_lower_sigma, "\n")

# %%
"""
All methods above are available as an `instance`:
"""

# %%
instance_at_upper_sigma = samples.instance_at_upper_sigma
instance_at_lower_sigma = samples.instance_at_lower_sigma
error_instance_at_upper_sigma = samples.error_instance_at_upper_sigma
error_instance_at_lower_sigma = samples.error_instance_at_lower_sigma

# %%
"""
An `instance` of any accepted parameter space sample can be created:
"""

# %%
instance = samples.instance_from_sample_index(sample_index=500)
print("Gaussian Instance of sample 5000:")
print("Centre = ", instance.profiles.gaussian.centre)
print("Intensity = ", instance.profiles.gaussian.intensity)
print("Sigma = ", instance.profiles.gaussian.sigma, "\n")

# %%
"""
Because `DynestyStatic`, a nested sampling *_NonLinearSearch_* was used, the `log_evidence` of the model is also 
available which enables Bayesian model comparison to be performed.
"""

# %%
log_evidence = samples.log_evidence

# %%
"""
Finally, lets remind ourselves of the `Result` class in the module:

 `chapter_2_results/src/phase/result.py` 
 
Here, we extended the `Result` class with two additional methods:

 - `max_log_likelihood_model_data`
 - `max_log_likelihood_fit`
"""

# %%
htf.plot.Line.line(xvalues=dataset.xvalues, line=result.max_log_likelihood_model_data)
htf.plot.FitDataset.chi_squared_map(fit=result.max_log_likelihood_fit)

# %%
"""
The Probability Density Functions (PDF`s) of the results can be plotted using the library:

 `corner.py`: https://corner.readthedocs.io/en/latest/

(In built visualization for PDF`s and `NonLinearSearch`'s is a future feature of **PyAutoFit**, but for now you`ll 
have to use the libraries yourself!).
"""

import corner

corner.corner(xs=samples.parameters, weights=samples.weights)



# %%
"""
Tutorial 1: Model Mapping
=========================

In this tutorial, we'll parameterize a simple model and use **PyAutoFit** to map its parameters to a model instance,
which we'll ultimately need to fit to data.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
import numpy as np
import matplotlib.pyplot as plt

# %%
"""
The tutorials need to know the path to your autofit_workspace folder, in order to:

 - Load configuration settings from the config files.
 - Load example data.
 - Output the results of models fits to your hard-disk. 

If you don't have an autofit_workspace (perhaps you cloned / forked the **PyAutoLens** GitHub repository?) you can
download it here:
 
 ttps://github.com/Jammy2211/autofit_workspace

Make sure to set up your WORKSPACE environment variable correctly, using either the "setup_environment.py" script 
supplied in the workspace or as described in the installation instructions:

https://pyautofit.readthedocs.io/en/latest/general/installation.html
    
This WORKSPACE environment variable is used in each tutorial to determine the path to the autofit_workspace, 
as shown below. 
"""

# %%
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
You`re going to see a line like the one below (with `conf.instance =`) in every tutorial this chapter. This sets the
following property:

 - The path to the configuration files used by **PyAutoFit**. You need to give the path to your autofit_workspace, so 
 the configuration files in the workspace are used (e.g. `/path/to/autofit_workspace/config`). 

(These will work autommatically if the WORKSPACE environment variable was set up correctly during installation. 
Nevertheless, setting the paths explicitly within the code is good practise.
"""

# %%
conf.instance = conf.Config(config_path=f"{workspace_path}/config")

# %%
"""
Below, you`ll notice the command:

 `from howtofit.simulators.chapter_1.gaussian_x1`

This will crop up in nearly every tutorial from here on. This imports a module that simulates the `Dataset` we plot in
this tutorialt. Feel free to check out the simulator scripts to see how this is done!

 - The data is a 1D numpy array of values corresponding to the observed counts of the Gaussian.
 - The noise-map corresponds to the expected noise in every data point.
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

data = gaussian_x1.data
noise_map = gaussian_x1.noise_map

# %%
"""
Lets plot the `Gaussian` using Matplotlib. 

The `Gaussian` is on a line of xvalues, which we'll compute using the shape of the `Gaussian` data and plot on the x-axis.
These xvalues will be used in later tutorials to create and fit Gaussians to the data.
"""

# %%
xvalues = np.arange(data.shape[0])
plt.plot(xvalues, data)
plt.show()

# %%
"""
Its not until tutorial 3 that we'll actually fit this image with a model. But its worth us looking at it now so we
can understand the model we`re going to fit. So what is the model?

Clearly, its a one-dimensional `Gaussian` defined as:

\begin{equation*}
g(x, I, \sigma) = \frac{I}{\sigma\sqrt{2\pi}} \exp{(-0.5 (x / \sigma)^2)}
\end{equation*}

Where:

x - Is x-axis coordinate where the `Gaussian` is evaluated.
I - Describes the intensity of the Gaussian.
sigma - Describes the size of the Gaussian.

This simple equation describes our model - a 1D `Gaussian` - and it has 3 parameters, $(x, I, \sigma)$. Using different
values of these 3 parameters we can describe *any* possible 1D Gaussian.

At its core, **PyAutoFit** is all about making it simple to define a model and straight forwardly map a set of input
parameters to the model.

So lets go ahead and create our model of a 1D Gaussian.
"""

# %%
class Gaussian:
    def __init__(
        self,
        centre=0.0,  # <- **PyAutoFit** recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian`s model parameters.
        sigma=0.01,
    ):
        self.centre = centre
        self.intensity = intensity
        self.sigma = sigma

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the light profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues : ndarray
            The x coordinates in the original reference frame of the data.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


# %%
"""
The class`s format is how **PyAutoFit** requires the components of a model to be written, where:

- The name of the class is the name of the model component, in this case, "Gaussian".

- The input arguments of the constructor are the model parameters which we will ultimately fit for, in this case the
  centre, intensity and sigma.
  
- The default values of the input arguments tell **PyAutoFit** whether a parameter is a single-valued floats or a 
  multi-valued tuple. For the `Gaussian` class, no input parameters are a tuple and we will show an example of a tuple 
  input in a later tutorial).
  
By writing a model component in this way, we can use the Python class to set it up as model component in **PyAutoFit**.
**PyAutoFit** can the generate model components as instances of their Python class, meaning that its functions 
(e.g. `profile_from_xvalues`) are accessible to **PyAutoFit**.

To set it up as a model component, we use a `PriorModel` object.
"""

# %%
model = af.PriorModel(Gaussian)
model.centre = af.UniformPrior(lower_limit=0.0, upper_limit=np.inf)
model.intensity = af.UniformPrior(lower_limit=0.0, upper_limit=np.inf)
model.sigma = af.UniformPrior(lower_limit=0.0, upper_limit=np.inf)
print("PriorModel `Gaussian` object: \n")
print(model)

# %%
"""
Using this `PriorModel` we can create an `instance` of the model, by mapping a list of physical values of each parameter 
as follows.
"""

# %%
instance = model.instance_from_vector(vector=[1.0, 2.0, 3.0])

# %%
"""
The instance is an instance of the `Gaussian` class.
"""

# %%
print("Model Instance: \n")
print(instance)

# %%
"""
It has the parameters of the `Gaussian` with the values input above.
"""

# %%
print("Instance Parameters \n")
print("x = ", instance.centre)
print("intensity = ", instance.intensity)
print("sigma = ", instance.sigma)

# %%
"""
Congratulations! You`ve defined your first model in **PyAutoFit**! :)

So, why is it called a PriorModel?

The parameters of a `PriorModel` in **PyAutoFit** all have a prior associated with them. Priors encode our expectations on
what values we expect each parameter can have. For example, we might know that our `Gaussian` will be centred near 0.0.

How are priors set? In this example, we did not specify priors, so they default to UniformPriors between 0 and 1. Below,
we'll show how to customize priors, and in a later tutorial we'll explain how for a given model component we can 
define config files that specify the default priors.

__How Are Priors Used?__

Priors are used to create model instances from a unit-vector, which is a vector defined in the same way as the vector 
above but with values spanning from 0 -> 1.

Unit values are mapped to physical values via the prior, for example:

For a UniformPrior defined between 0.0 and 10.0:

- An input unit value of 0.5 will give the physical value 5.0.
- An input unit value of 0.8 will give te physical value 8.0.

For a LogUniformPrior (base 10) defined between 1.0 and 100.0:

- An input unit value of 0.5 will give the physical value 10.0.
- An input unit value of 1.0 will give te physical value 100.0.

For a GauassianPrior defined with mean 1.0 and sigma 1.0:

- An input unit value of 0.5 (e.g. the centre of the Gaussian) will give the physical value 1.0.
- An input unit value of 0.8173 (e.g. 1 sigma confidence) will give te physical value 1.9051.

Lets take a look:
"""

# %%
"""
We can overwrite the default priors assumed for each parameter.
"""

# %%
model.centre = af.UniformPrior(lower_limit=10.0, upper_limit=20.0)
model.intensity = af.GaussianPrior(mean=5.0, sigma=7.0)
model.sigma = af.LogUniformPrior(lower_limit=1.0, upper_limit=100.0)

# %%
"""
These priors are now used to map our unit values to physical values when we create an instance of the Gaussian
class.
"""

# %%
instance = model.instance_from_unit_vector(unit_vector=[0.5, 0.3, 0.8])

# %%
"""
Lets check that this instance is again an instance of the `Gaussian` class.
"""

# %%
print("Model Instance: \n")
print(instance)

# %%
"""
It now has physical values for the parameters mapped from the priors defined above.
"""

# %%
print("Instance Parameters \n")
print("x = ", instance.centre)
print("intensity = ", instance.intensity)
print("sigma = ", instance.sigma)

# %%
"""
We can also set physical limits on parameters, such that a model instance cannot generate parameters outside of a
specified range.

For example, a `Gaussian` cannot have a negative intensity, so we can set its lower limit to a value of 0.0.
"""

# %%
model.intensity = af.GaussianPrior(
    mean=0.0, sigma=1.0, lower_limit=0.0, upper_limit=1000.0
)

# %%
"""
The unit vector input below creates a negative intensity value, such that if you uncomment the line below **PyAutoFit** 
raises an error.
"""

# %%
# instance = model.instance_from_unit_vector(unit_vector=[0.01, 0.01, 0.01])

# %%
"""
In a later tutorial, we'll explain how config files can again be used to set the default limits of every parameter.


And with that, you`ve completed tutorial 1!

At this point, you might be wondering, whats the big deal? Sure, its cool that we set up a model and its nice that
we can translate priors to parameters in this way, but how is this actually going to help me perform model fitting?
With a bit of effort couldn`t I have written some code to do this myself?

Well, you`re probably right, but this tutorial is covering just the backend of **PyAutoFit** - what holds everything
together. Once you start using **PyAutoFit**, its unlikely that you`ll perform model mapping yourself, its the `magic` 
behind the scenes that makes model-fitting work.

So, we`re pretty much ready to move on to tutorial 2, where we'll actually fit this model to some data. However,
first, I want you to quickly think about the model you want to fit. How would you write it as a class using the
**PyAutoFit** format above? What are the free parameters of you model? Are there multiple model components you are going
to want to fit to your data?

Below are two more classes one might use to perform model fitting, the first is the model of a linear-regression line
of the form $y = mx + c$ that you might fit to a 1D data-set:
"""

# %%
class LinearFit:
    def __init__(self, gradient=1.0, intercept=0.0):

        self.gradient = gradient
        self.intercept = intercept


# %%
"""
The second example is a two-dimensional Gaussian. Here, the centre now has two coordinates (y,x), which in **PyAutoFit**
is more suitably defined using a tuple.
"""

# %%
class Gaussian2D:
    def __init__(self, centre=(0.0, 0.0), intensity=0.1, sigma=1.0):

        self.centre = centre
        self.intensity = intensity
        self.sigma = sigma



# %%
"""
Tutorial 3: Aggregator
======================

In the previous tutorial, we fitted 3 `Datasets` with an identical `Phase`, outputting the results of each to a unique
folder on our hard disk.

However, lets use the `Aggregator` to load the `Result`'s and manipulate / plot them using our Jupyter notebook. The API
for using `Result`'s follow closely tutorial 1 of this chapter.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_2_results import src as htf

import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_2",
)

# %%
"""
To load the results of the previous tutorial into the `Aggregator`, we simply point the `Aggregator` class to the path 
of the results we want it to load.
"""

# %%
output_path = f"{workspace_path}/output/chapter_2/aggregator"

agg = af.Aggregator(directory=str(output_path))

# %%
"""
To begin, let me quickly explain what a generator is in Python, for those unaware. A generator is an object that 
iterates over a function when it is called. The `Aggregator` creates all objects as generators, rather than lists, or 
dictionaries, or whatever.

Why? Because lists and dictionaries store every entry in memory simultaneously. If you fit many `Dataset`s, you`ll 
have lots of results and therefore use a lot of memory. This will crash your laptop! On the other hand, a generator 
only stores the object in memory when it runs the function; it is free to overwrite it afterwards. Thus, your laptop 
won`t crash!

There are two things to bare in mind with generators:

 1) A generator has no length, thus to determine how many entries of data it corresponds to you first must turn it to a 
    list.

 2) Once we use a generator, we cannot use it again and we'll need to remake it. For this reason, we typically avoid 
    storing the generator as a variable and instead use the `Aggregator` to create them on use.

We can now create a `Samples` generator of every fit. This creates instances of the `Samples` class we manipulated in
tutorial 1, which with the `Aggregator` now acts as an interface between the results of the non-linear fit on your 
hard-disk and Python.
"""

# %%
samples_gen = agg.values("samples")

# %%
"""
When we print this list of outputs you should see over 3 different `NestSamples` instances, corresponding to the 3
model-fits we performed in the previous tutorial.
"""

# %%
print("Emcee Samples:\n")
print(samples_gen)
print("Total Samples Objects = ", len(list(samples_gen)), "\n")


# %%
"""
We've encountered the `Samples` class in previous tutorials. As we saw in tutorial 1, the `Samples` class contains all 
the accepted parameter samples of the `NonLinearSearch`, which is a list of lists where:

 - The outer list is the size of the total number of samples.
 - The inner list is the size of the number of free parameters in the fit.

With the `Aggregator` we can now get information on the `Samples` of all 3 model-fits, as opposed to just 1 fit using 
its `Result` object.
"""

# %%
for samples in agg.values("samples"):
    print("All parameters of the very first sample")
    print(samples.parameters[0])
    print("The tenth sample`s third parameter")
    print(samples.parameters[9][2])
    print()

# %%
"""
We can use the `Aggregator` to get information on the `log_likelihoods`, log_priors`, `weights`, etc. of every fit.
"""

# %%
for samples in agg.values("samples"):
    print("log(likelihood), log(prior), log(posterior) and weight of the tenth sample.")
    print(samples.log_likelihoods[9])
    print(samples.log_priors[9])
    print(samples.log_posteriors[9])
    print(samples.weights[9])
    print()

# %%
"""
We can use the `Sample`'s to create a list of the `max_log_likelihood_vector` of each fit to our three images.
"""

# %%
vector = [samps.max_log_likelihood_vector for samps in agg.values("samples")]
print("Maximum Log Likelihood Parameter Lists:\n")
print(vector, "\n")

# %%
"""
As discussed in tutorial 1, using vectors isn't too much use, as we can`t be sure which values correspond to which 
parameters.

We can use the `Aggregator` to create the `max_log_likelihood_instance` of every fit.
"""

# %%
instances = [samps.max_log_likelihood_instance for samps in agg.values("samples")]
print("Maximum Log Likelihood Model Instances:\n")
print(instances, "\n")

# %%
"""
The model instance contains all the model components of our fit which for the fits above was a single `Gaussian`
profile (the word `gaussian` comes from what we called it in the `CollectionPriorModel` above).
"""

# %%
print(instances[0].profiles.gaussian)
print(instances[1].profiles.gaussian)
print(instances[2].profiles.gaussian)

# %%
"""
This, of course, gives us access to any individual parameter of our maximum log likelihood `instance`. Below, we see 
that the 3 `Gaussian`s were simulated using `sigma` values of 1.0, 5.0 and 10.0.
"""

# %%
print(instances[0].profiles.gaussian.sigma)
print(instances[1].profiles.gaussian.sigma)
print(instances[2].profiles.gaussian.sigma)

# %%
"""
We can also access the `median_pdf` model via the `Aggregator`, as we saw for the `Samples` object in tutorial 1.
"""

# %%
mp_vectors = [samps.median_pdf_vector for samps in agg.values("samples")]
mp_instances = [samps.median_pdf_instance for samps in agg.values("samples")]

print("Median PDF Model Parameter Lists:\n")
print(mp_vectors, "\n")
print("Most probable Model Instances:\n")
print(mp_instances, "\n")

# %%
"""
We can also print the `model_results` of all phases, which is string that summarizes every fit`s model providing
quick inspection of all results.
"""

# %%
results = agg.model_results
print("Model Results Summary:\n")
print(results, "\n")

# %%
"""
Lets end the tutorial with something more ambitious. Lets create a plot of the inferred `sigma` values vs `intensity` 
of each `Gaussian` profile, including error bars at $3\sigma$ confidence.
"""

# %%
import matplotlib.pyplot as plt

mp_instances = [samps.median_pdf_instance for samps in agg.values("samples")]
ue3_instances = [
    samp.error_instance_at_upper_sigma(sigma=3.0) for samp in agg.values("samples")
]
le3_instances = [
    samp.error_instance_at_lower_sigma(sigma=3.0) for samp in agg.values("samples")
]

mp_sigmas = [instance.profiles.gaussian.sigma for instance in mp_instances]
ue3_sigmas = [instance.profiles.gaussian.sigma for instance in ue3_instances]
le3_sigmas = [instance.profiles.gaussian.sigma for instance in le3_instances]
mp_intensitys = [instance.profiles.gaussian.sigma for instance in mp_instances]
ue3_intensitys = [instance.profiles.gaussian.sigma for instance in ue3_instances]
le3_intensitys = [instance.profiles.gaussian.intensity for instance in le3_instances]

plt.errorbar(
    x=mp_sigmas,
    y=mp_intensitys,
    marker=".",
    linestyle="",
    xerr=[le3_sigmas, ue3_sigmas],
    yerr=[le3_intensitys, ue3_intensitys],
)
plt.show()



# %%
"""
Tutorial 5: Data and Models
===========================

Up to now, we've used used the `Aggregator` to load and inspect the `Samples` of 3 model-fits.

In this tutorial, we'll look at how the way we designed our source code makes it easy to use the `Aggregator` to
inspect, interpret and plot the results of the model-fit, including refitting the best models to our data.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_2_results import src as htf

import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_2",
)

# %%
"""
To load these results with the `Aggregator`, we again point it to the path of the results we want it to inspect, with
our path straight to the `Aggregator` results ensuring we don't need to filter our `Aggregator` in this tutorial.
"""

# %%
output_path = f"{workspace_path}/output/chapter_2/aggregator"
agg = af.Aggregator(directory=str(output_path))

# %%
"""
We can use the `Aggregator` to load a generator of every fit`s dataset, by changing the `output` attribute to the 
`dataset` attribute at the end of the aggregator.

Note that in the source code for chapter 2, specifically in the `phase.py` module, we specified that the the `Dataset` 
object would be saved to hard-disk such that the `Aggregator` can load it.
"""

# %%
dataset_gen = agg.values("dataset")
print("Datasets:")
print(list(dataset_gen), "\n")

# %%
"""
It is here the object-oriented design of our `plot.py` module comes into its own. We have the `Dataset` objects loaded, 
meaning we can easily plot each `Dataset` using the `dataset_plot.py` module.
"""

# %%
for dataset in agg.values("dataset"):
    htf.plot.Dataset.data(dataset=dataset)

# %%
"""
The `Dataset` names are available as part of the `Dataset`.
"""

# %%
for dataset in agg.values("dataset"):
    print(dataset.name)

# %%
"""
The `info` dictionary we input into the `Phase` is also available.
"""

# %%
for info in agg.values("info"):
    print(info)

# %%
"""
We can repeat the same trick to get the `mask` of every fit.
"""

# %%
mask_gen = agg.values("mask")
print("Masks:")
print(list(mask_gen), "\n")


# %%
"""
We`re going to refit each `Dataset` with the `max_log_likelihood_instance` of each model-fit. To do this, we'll need 
each `Phase`'s `MaskedDataset`.

(If you are unsure what the `zip` is doing below, it essentially combines the `dataset_gen`, `mask_gen` and 
`settings_gen` into one list such that we can iterate over all three simultaneously to create each `MaskedDataset`).

The `MaskedDataset` may have been altered by the `data_trim_left` and `data_trim_right` `SettingsPhase`. We can 
load the `SettingsPhase` via the `Aggregator` to use these settings when we create the `MaskedDataset`.
"""

# %%
dataset_gen = agg.values("dataset")
mask_gen = agg.values("mask")
settings_gen = agg.values("settings")

masked_datasets = [
    htf.MaskedDataset(
        dataset=dataset, mask=mask, settings=settings.settings_masked_dataset
    )
    for dataset, mask, settings in zip(dataset_gen, mask_gen, settings_gen)
]

# %%
"""
There is a problem with how we set up the `MaskedDataset`s above, can you guess what it is?

We used lists! If we had fit a large sample of data, the above object would store the `MaskedDataset` of all objects
simultaneously in memory on our hard-disk, likely crashing our laptop! To avoid this, we must write functions that
manipulate the `Aggregator` generators as generators themselves. Below is an example function that performs the same
task as above.
"""

# %%
def masked_dataset_from_agg_obj(agg_obj):

    dataset = agg_obj.dataset
    mask = agg_obj.mask
    settings = agg_obj.settings

    return htf.MaskedDataset(
        dataset=dataset, mask=mask, settings=settings.settings_masked_dataset
    )


# %%
"""
To manipulate this function as a generator using the `Aggregator`, we apply it to the `Aggregator`'s `map` function.

The `masked_dataset_gen` below ensures that we avoid representing all `MaskedDataset`'s simultaneously in memory.
"""

# %%
masked_dataset_gen = agg.map(func=masked_dataset_from_agg_obj)
print(list(masked_dataset_gen))

# %%
"""
Lets get the `max_log_likelihood_instance`s, as we did in tutorial 3.
"""

# %%
instances = [samps.max_log_likelihood_instance for samps in agg.values("samples")]

# %%
"""
Okay, we want to inspect the fit of each `max_log_likelihood_instance`. To do this, we reperform each fit.

First, we need to create the `model_data` of every `max_log_likelihood_instance`. Lets begin by creating a list 
of profiles of every phase.
"""

# %%
profiles = [instance.profiles for instance in instances]

# %%
"""
We can use these to create the `model_data` of each set of profiles (which in this case is just 1 `Gaussian`, but had
we included more profiles in the model would consist of multiple `Gaussian`s / `Exponential`s.).
"""

# %%
model_datas = [
    profile.gaussian.profile_from_xvalues(xvalues=dataset.xvalues)
    for profile, dataset in zip(profiles, agg.values("dataset"))
]

# %%
"""
And, as we did in tutorial 2, we can combine the `MaskedDataset`s and `model_data`s in a `Fit` object to create the
maximum likelihood fit of each phase!
"""

# %%
fits = [
    htf.FitDataset(masked_dataset=masked_dataset, model_data=model_data)
    for masked_dataset, model_data in zip(masked_datasets, model_datas)
]

# %%
"""
We can now plot different components of the `Fit` (again benefiting from how we set up the `fit_plots.py` module)!
"""

# %%
for fit in fits:
    htf.plot.FitDataset.residual_map(fit=fit)
    htf.plot.FitDataset.normalized_residual_map(fit=fit)
    htf.plot.FitDataset.chi_squared_map(fit=fit)

# %%
"""
Again, the code above does not use generators and could prove memory intensive for large datasets. Below is how we 
would perform the above task with generator functions, using the `masked_dataset_gen` above for the `MaskedDataset`.
"""

# %%
def model_data_from_agg_obj(agg_obj):

    xvalues = agg_obj.dataset.xvalues
    instance = agg_obj.samples.max_log_likelihood_instance
    profiles = instance.profiles

    return sum([profile.profile_from_xvalues(xvalues=xvalues) for profile in profiles])


def fit_from_agg_obj(agg_obj):

    masked_dataset = masked_dataset_from_agg_obj(agg_obj=agg_obj)
    model_data = model_data_from_agg_obj(agg_obj=agg_obj)

    return htf.FitDataset(masked_dataset=masked_dataset, model_data=model_data)


fit_gen = agg.map(func=fit_from_agg_obj)

for fit in fit_gen:
    htf.plot.FitDataset.residual_map(fit=fit)
    htf.plot.FitDataset.normalized_residual_map(fit=fit)
    htf.plot.FitDataset.chi_squared_map(fit=fit)

# %%
"""
Setting up the above objects (the `masked_dataset`s, `model data`s, `fit`s) was a bit of work. It wasn`t too many 
lines of code, but for something our users will want to do many times it`d be nice to have a short cut to setting them 
up, right?

In the source code module `aggregator.py` we've set up exactly such a short-cut. This module simply contains the 
generator functions above such that the generator can be created by passing the `Aggregator`. This provides us with 
convenience methods for quickly creating the `MaskedDataset`, `model_data` and `Fit`'s using a single line of code:
"""

# %%

masked_dataset_gen = htf.agg.masked_dataset_generator_from_aggregator(aggregator=agg)
model_data_gen = htf.agg.model_data_generator_from_aggregator(aggregator=agg)
fit_gen = htf.agg.fit_generator_from_aggregator(aggregator=agg)

htf.plot.FitDataset.residual_map(fit=list(fit_gen)[0])

# %%
"""
The methods in `aggregator.py` actually allow us to go one step further: they all us to create the `MaskedDataset` and
`Fit` objects using an input `SettingsMaskedDataset`. This means we can fit a `Dataset` with a `Phase` and then see how
the model-fits change if we customize the `Dataset` in different ways.

Below, we create and plot a `Fit` where the `MaskedDataset` is trimmed from the left and right.
"""

# %%
settings_masked_dataset = htf.SettingsMaskedDataset(
    data_trim_left=20, data_trim_right=20
)

fit_gen = htf.agg.fit_generator_from_aggregator(
    aggregator=agg, settings_masked_dataset=settings_masked_dataset
)

htf.plot.FitDataset.residual_map(fit=list(fit_gen)[0])

# %%
"""
For your model-fitting project, you`ll need to update the `aggregator.py` module in the same way. This is why we have 
emphasised the object-oriented design of our model-fitting project throughout. This design makes it very easy to 
inspect results via the `Aggregator` later on!
"""



import pytest
from scipy import stats

import autofit.mapper.variable
from autofit import graphical as mp


@pytest.fixture(
    name="x"
)
def make_x():
    return autofit.mapper.variable.Variable("x")


@pytest.fixture(
    name="probit_factor"
)
def make_probit_factor(x):
    return mp.Factor(
        stats.norm(
            loc=0.,
            scale=1.
        ).logcdf,
        x=x
    )



# %%
"""
Tutorial 3: Non-linear Search
=============================

Okay, so its finally time to take our model and fit it to data, hurrah!

So, how do we infer the parameters for the 1D `Gaussian` that give a good fit to our data?  In the last tutorial, we
tried a very basic approach, randomly guessing models until we found one that gave a good fit and high log_likelihood.

We discussed that this wasn`t really a viable strategy for more complex models, and it isn't. However, this is the
basis of how model fitting actually works! Basically, our model-fitting algorithm guesses lots of models, tracking
the log likelihood of these models. As the algorithm progresses, it begins to guess more models using parameter
combinations that gave higher log_likelihood solutions previously. If a set of parameters provided a good fit to the
data previously, a model with similar values probably will too.

This is called a `_NonLinearSearch_` and its a fairly common problem faced by scientists. We`re going to use a
_NonLinearSearch_ algorithm called `Emcee`, which for those familiar with statistic inference is a Markov Chain Monte
Carlo (MCMC) method. For now, lets not worry about the details of how Emcee actually works. Instead, just picture that
a `NonLinearSearch` in **PyAutoFit** operates as follows:

 1) Randomly guess a model, mapping their parameters via the priors to instances of the model, in this case a
 Gaussian.

 2) Use this model instance to generate model data and compare this model data to the data to compute a log
 likelihood.

 3) Repeat this many times, choosing models whose parameter values are near those of the model which currently has
 the highest log likelihood value. If the new model`s log likelihood is higher than the previous model, new
 models will thus be chosen with parameters nearer this model.

The idea is that if we keep guessing models with higher log-likelihood values, we'll inevitably `climb` up the gradient
of the log likelihood in parameter space until we eventually hit the highest log likelihood models.

To be clear, this overly simplfied description of an MCMC algorithm is not how the *_NonLinearSearch_* really works and
omits the details of how our priors inpact our inference or why an MCMC algorithm can provide reliable errors on our
parameter estimates.

A detailed understanding of *non-linear sampling* is not required in chapter 1, however interested users will find
a complete description of *non-linear searches* in chapter 2 of the HowToFit lectures.
"""

# %%
#%matplotlib inline

# %%
from autoconf import conf
import autofit as af
import matplotlib.pyplot as plt
import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""

The line `conf.instance` is now used to set up a second property of the configuration; the path to the **PyAutoFit** 
output folder, which is where the results of the `NonLinearSearch` are written to on your hard-disk, alongside 
visualization and other properties of the fit (e.g. `/path/to/autolens_workspace/output/howtolens`)

(These will work autommatically if the WORKSPACE environment variable was set up correctly during installation. 
Nevertheless, setting the paths explicitly within the code is good practise.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_1",  # <- This sets up where the `NonLinearSearch`'s outputs go.
)

# %%
"""
To create the `Dataset`, we import the simulator module and use it to generate the `Dataset`'s data and noise-map. 
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

data = gaussian_x1.data
noise_map = gaussian_x1.noise_map

# %%
"""
Lets remind ourselves what the data looks like, using the `plot_line` convenience method fom the previous tutorial.
"""

# %%
def plot_line(xvalues, line, ylabel=None):

    plt.plot(xvalues, line)
    plt.xlabel("x")
    plt.ylabel(ylabel)
    plt.show()
    plt.clf()


xvalues = np.arange(data.shape[0])

plot_line(xvalues=xvalues, line=data, ylabel="Data")

# %%
"""
Lets remake the `Gaussian` class for this tutorial, which is the model we will fit using the `NonLinearSearch`.
"""

# %%
class Gaussian:
    def __init__(
        self,
        centre=0.0,  # <- **PyAutoFit** recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian`s model parameters.
        sigma=0.01,
    ):
        self.centre = centre
        self.intensity = intensity
        self.sigma = sigma

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the light profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues : ndarray
            The x coordinates in the original reference frame of the data.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


# %%
"""
The `NonLinearSearch` requires an `Analysis` class, which:

 - Receives the data to be fitted and prepares it so the model can fit it.
 
 - Defines the `log_likelihood_function` used to compute the log likelihood from a model instance. 
 
 - Passes this log likelihood to the `NonLinearSearch` so that it can determine parameter values for the the next model 
 that it samples.

For our 1D `Gaussian` model-fitting example, here is our `Analysis` class:
"""

# %%
class Analysis(af.Analysis):
    def __init__(self, data, noise_map):

        super().__init__()

        self.data = data
        self.noise_map = noise_map

    def log_likelihood_function(self, instance):

        """
        The `instance` that comes into this method is an instance of the `Gaussian` class above, with the parameters
        set to values chosen by the `NonLinearSearch`. (These are commented out to pretent excessive print statements
        when we run the `NonLinearSearch`..

        This instance`s parameter values are chosen by the `NonLinearSearch` based on the previous model with the
        highest likelihood result.

            print("Gaussian Instance:")
            print("Centre = ", instance.centre)
            print("Intensity = ", instance.intensity)
            print("Sigma = ", instance.sigma)

        Below, we fit the data with the `Gaussian` instance, using its "profile_from_xvalues" function to create the
        model data.
        """

        xvalues = np.arange(self.data.shape[0])

        model_data = instance.profile_from_xvalues(xvalues=xvalues)
        residual_map = self.data - model_data
        chi_squared_map = (residual_map / self.noise_map) ** 2.0
        chi_squared = sum(chi_squared_map)
        noise_normalization = np.sum(np.log(2 * np.pi * noise_map ** 2.0))
        log_likelihood = -0.5 * (chi_squared + noise_normalization)

        return log_likelihood


# %%
"""
To perform the `NonLinearSearch` using `Emcee`, we simply compose our model using a `PriorModel`, instantiate the 
_Analysis_ class and pass them to an instance of the `Emcee` class.

We manually set the priors on the model, in the next tutorial we'll cover how this can be performed automatically.
"""

# %%
model = af.PriorModel(Gaussian)
model.centre = af.UniformPrior(lower_limit=0.0, upper_limit=100.0)
model.intensity = af.UniformPrior(lower_limit=0.0, upper_limit=1e2)
model.sigma = af.UniformPrior(lower_limit=0.0, upper_limit=50.0)

analysis = Analysis(data=data, noise_map=noise_map)

emcee = af.Emcee()

# %%
"""
We begin the `NonLinearSearch` by calling its `fit` method, it  will take a minute or so to run (which is very fast 
for a model-fit). Whilst you`re waiting, checkout the folder:

`autofit_workspace/howtofit/chapter_1_introduction/output/emcee`

Here, the results of the model-fit are output to your hard-disk (on-the-fly) and you can inspect them as the non-linear
search runs. In particular, you`ll find:

 - `model.info`: A file listing every model component, parameter and prior in your model-fit.

 - `model.results`: A file giving the latest best-fit model, parameter estimates and errors of the fit.
 
 - `search`: A folder containing the `Emcee` output in hdf5 format.txt (you`ll probably never need to look at these, 
   but its good to know what they are).
 
 - Other `metadata` which you can ignore for now.
"""

# %%
result = emcee.fit(model=model, analysis=analysis)

print(
    "Emcee has begun running - checkout the autofit_workspace/howtofit/chapter_1_introduction/output/phase_t3"
    " folder for live output of the results."
    "This Jupyter notebook cell with progress once Emcee has completed - this could take a few minutes!"
)

print("Emcee has finished run - you may now continue the notebook.")

# %%
"""
Once completed, the `NonLinearSearch` returns a `Result` object, which contains lots of information about the 
_NonLinearSearch_.
 
A full description of the `Results` object can be found at:
 
`autofit_workspace/examples/simple/results`
`autofit_workspace/examples/complex/results`.

In this tutorial, lets use it to inspect the maximum likelihood model instance.
"""

# %%
print("Maximum Likelihood Model:\n")
max_log_likelihood_instance = result.samples.max_log_likelihood_instance
print("Centre = ", max_log_likelihood_instance.centre)
print("Intensity = ", max_log_likelihood_instance.intensity)
print("Sigma = ", max_log_likelihood_instance.sigma)

# %%
"""
We can use this to plot the maximum log likelihood fit over the data:
"""

# %%
model_data = result.max_log_likelihood_instance.profile_from_xvalues(
    xvalues=np.arange(data.shape[0])
)
plt.plot(range(data.shape[0]), data)
plt.plot(range(data.shape[0]), model_data)
plt.xlabel("x")
plt.ylabel("Intensity")
plt.show()
plt.close()

# %%
"""
Above, we used the `Result`'s `samples` property, which in this case is a `MCMCSamples` object:
"""

# %%
print(result.samples)

# %%
"""
This object acts as an interface between the `Emcee` output results on your hard-disk and this Python code. For
example, we can use it to get the parameters and log likelihood of every accepted emcee sample.
"""

# %%
print(result.samples.parameters)
print(result.samples.log_likelihoods)

# %%
"""
We can also use it to get a model instance of the `median_pdf` model, which is the model where each parameter is
the value estimated from the probability distribution of parameter space.
"""

# %%
mp_instance = result.samples.median_pdf_instance
print()
print("Median PDF Model:\n")
print("Centre = ", mp_instance.centre)
print("Intensity = ", mp_instance.intensity)
print("Sigma = ", mp_instance.sigma)

# %%
"""
we'll come back to look at `Samples` objects in more detail in chapter 2!
"""



# %%
"""
Tutorial 4: Filtering
=====================

In the previous tutorials, the `Aggregator` has always loaded all of the results of all 3 fits.

However, suppose we had the results of other fits in the output folder and we *only* wanted fits which used
a certain phase. Or, imagine we want the results of a fit to 1 specific data. In this tutorial, we'll learn how to use
the `Aggregator`'s `filter` tool, which filters the results and provides us with only the results we want.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_2_results import src as htf

import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs and output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output",
)

# %%
"""
When we load the `Aggregator`, in this tutorial the `output_path` does not include the `/chapter2/aggregator` on the 
end, like it did in the previous tutorial. 

This means that, without filters, the `Aggregator` will load all results in the output folder, including those from 
chapter 1 and the previous tutorials in this chapter.

We can first filter results to only include completed results. By including `completed_only` below, any results which 
are in the middle of a `NonLinearSearch` will be omitted and not loaded in the `Aggregator`.
"""

# %%
output_path = f"{workspace_path}/output"

agg = af.Aggregator(directory=str(output_path), completed_only=True)

# %%
"""
First, lets print the number of `Samples` objects the `Aggregator` finds. Now we are pointing the `Aggregator` to the
output folder, depending on what results are there this should find in excess of 3 sets of results, corresponding to
results from chapter 1 and tutorial 1 of chapter 2.
"""

# %%
print("Emcee Samples:\n")
print("Total Samples Objects = ", len(list(agg.values("samples"))), "\n")

# %%
"""
To remove the fits of previous tutorials and just keep the `MCMCSamples` of the 3 `Dataset`s fitted in this tutorial 
we can use the `Aggregator`'s `filter` tool. Below, we use the `phase_name` of the results, `phase_t2` which is 
unique to all 3  fits, to filter our results as desired.
"""

# %%
phase_name = "phase_t2_agg"
agg_filter = agg.filter(agg.phase == phase_name)

# %%
"""
This filtered `Aggregator` should now produce just 3 `Samples` objects.
"""
print("Total Samples Objects = ", len(list(agg.values("samples"))), "\n")
samples_gen = agg_filter.values("samples")

# %%
"""
Alternatively, we can filter using strings, requiring that the string appears in the full path of the output
results. This is useful if you fit a large sample of data where:

 - Multiple results, corresponding to different phases and model-fits are stored in the same path.
 
 - Different runs using different `SettingsPhase` are in the same path.
 
 - Fits using different `NonLinearSearch`s, with different settings, are contained in the same path.

The example below shows us using the `contains` method to filter the results of the fit to only the first `Dataset`. 
"""

# %%
agg_filter_contains = agg.filter(
    agg.directory.contains("phase_t2_agg"), agg.directory.contains("gaussian_x1_0")
)
print("Directory Contains Filtered NestedSampler Samples: \n")
print(
    "Total Samples Objects = ", len(list(agg_filter_contains.values("samples"))), "\n\n"
)

# %%
"""
You can also filter by the type of `NonLinearSearch` used, so that if you were to fit the same model to the same 
dataset with different `NonLinearSearch`s you could load and compare their results.
"""

# %%
agg_filter_contains = agg.filter(agg.non_linear_search == "dynestystatic")
print("Directory Contains Filtered NestedSampler Samples: \n")
print(
    "Total Samples Objects = ", len(list(agg_filter_contains.values("samples"))), "\n\n"
)

# %%
"""
Filters can be combined to load precisely only the result that you want, below we use all the above filters to 
load only the results of the fit to the first lens in our sample.
"""
agg_filter_multiple = agg.filter(
    agg.phase == "phase_t2_agg", agg.directory.contains("gaussian_x1_0")
)
print("Multiple Filter NestedSampler Samples: \n")
print()
print(
    "Total Samples Objects = ", len(list(agg_filter_multiple.values("samples"))), "\n\n"
)



# %%
"""
Tutorial 2: Dataset Sample
==========================

In this tutorial, we'll fit a multiple `Dataset`s with the same phase, producing multiple sets of results on our
hard-disk. In the following tutorials, we then use these results and the `Aggregator` to load the results into
our Jupyter notebook to interpret, inspect and plot the output results.

we'll fit 3 different `Dataset`s, each with a single `Gaussian` model.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
from howtofit.chapter_2_results import src as htf

import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
Setup the configs as we did in the previous tutorial, as well as the output folder for our `NonLinearSearch`.
"""

# %%
conf.instance = conf.Config(
    config_path=f"{workspace_path}/config",
    output_path=f"{workspace_path}/output/chapter_2",
)

# %%
"""
Here, for each `Dataset` we are going to set up the correct path, load it, create its `mask` and fit it using a `Phase`.

We want our results to be in a folder specific to the `Dataset`. we'll use the `Dataset`'s name string to do this. Lets
create a list of all 3 of our `Dataset` names.

we'll also pass these names to the `Dataset` when we create it, the name will be used by the `Aggregator` to name the 
file the data is stored. More importantly, the name will be accessible to the `Aggregator`, and we will use it to label 
figures we make via the `Aggregator`.
"""

# %%
from howtofit.simulators.chapter_2 import (
    gaussian_x1_0,
    gaussian_x1_1,
    gaussian_x1_2,
)

dataset_names = ["gaussian_x1_0", "gaussian_x1_1", "gaussian_x1_2"]
datas = [gaussian_x1_0.data, gaussian_x1_1.data, gaussian_x1_2.data]
noise_maps = [gaussian_x1_0.noise_map, gaussian_x1_1.noise_map, gaussian_x1_2.noise_map]

# %%
"""
We can also attach information to the model-fit, by setting up an info dictionary. 

Information about our model-fit (e.g. the data of observation) that isn't part of the model-fit is made accessible to 
the `Aggregator`. For example, below we write info on the `Dataset`'s data of observation and exposure time.
"""

# %%
info = {"date_of_observation": "01-02-18", "exposure_time": 1000.0}

# %%
"""
This for loop runs over every `Dataset`, checkout the comments below for how we set up the path structure.
"""

# %%
for index in range(len(datas)):

    """The code below creates the `Dataset` and `mask` as per usual."""

    dataset = htf.Dataset(data=datas[index], noise_map=noise_maps[index])

    mask = np.full(fill_value=False, shape=dataset.data.shape)

    """
    Here, we create a `Phase` as normal. However, we also include an input parameter `folders`. This defines the 
    names of any folders that the `Phase` outputs results to. For example, if a `Phase` outputs to the path:

        `/path/to/autofit_workspace/output/phase_name/`

    A `Phase` with the input `folders=[`phase_folder`]` edits this path to:

        `/path/to/autofit_workspace/output/phase_folder/phase_name/`

    You can input multiple `phase folders`, for example `folders=[`folder_0`, `folder_1`]` would create the path:

        `/path/to/autofit_workspace/output/folder_0/folder_1/phase_name/`

    Below, we use the `data_name`, so our results go in a folder specific to the `Dataset`, e.g:

        `/path/to/autofit_workspace/output/gaussian_x1_0/phase_t2/`
    """

    print(
        f"Emcee has begun running - checkout the "
        f"autofit_workspace/howtofit/chapter_2_results/output/aggregator/{dataset_names[index]}/phase_t2 folder for live "
        f"output of the results. This Jupyter notebook cell with progress once Emcee has completed - this could take a "
        f"few minutes!"
    )

    phase = htf.Phase(
        phase_name="phase_t2_agg",
        folders=["aggregator", dataset_names[index]],
        profiles=af.CollectionPriorModel(gaussian=htf.profiles.Gaussian),
        settings=htf.SettingsPhase(),
        search=af.DynestyStatic(),
    )

    """Note that we pass the info to the `phase` when we run it, so that the `Aggregator` can make it accessible."""

    phase.run(dataset=dataset, mask=mask, info=info)

# %%
"""
Checkout the output folder - you should see three new sets of results corresponding to our 3 `Gaussian` datasets.

Unlike previous tutorials, these folders in the output folder are named after the `Dataset` and contain the folder
with the phase`s name, as opposed to just the phase-name folder.
"""



# %%
"""
Tutorial 2: Model Fitting
=========================

In this tutorial, we'll fit the 1D `Gaussian` model from the previous tutorial to the data we loaded.
"""

# %%
#%matplotlib inline

from autoconf import conf
import autofit as af
import matplotlib.pyplot as plt
import numpy as np
import os

workspace_path = os.environ["WORKSPACE"]
print("Workspace Path: ", workspace_path)

# %%
"""
These setup the configs as we did in the previous tutorial.
"""

# %%
conf.instance = conf.Config(config_path=f"{workspace_path}/config")

# %%
"""
To create the Dataset, we import the simulator module and use it to generate the Dataset`s data and noise-map. 
"""

# %%
from howtofit.simulators.chapter_1 import gaussian_x1

data = gaussian_x1.data
noise_map = gaussian_x1.noise_map

# %%
"""
The shape of the data gives us its xvalues - the x coordinates we evaluate our model 1D `Gaussian` on.
"""

# %%
xvalues = np.arange(data.shape[0])
print(xvalues)

# %%
"""
Lets plot the data and noise-map we`re going to fit.
"""

# %%
plt.plot(xvalues, data)
plt.xlabel("x")
plt.ylabel("Intensity")
plt.show()
plt.clf()

plt.plot(xvalues, noise_map)
plt.xlabel("x")
plt.ylabel("Intensity")
plt.show()
plt.clf()

# %%
"""
So, how do we actually go about fitting our `Gaussian` model to this data? First, we need to be able to generate
an image of our 1D `Gaussian` model. As we did in tutorial 1, we define the `Gaussian` as a Python class with the format 
required for **PyAutoFit** to use it as a model-component.
"""

# %%
class Gaussian:
    def __init__(
        self,
        centre=0.0,  # <- **PyAutoFit** recognises these constructor arguments
        intensity=0.1,  # <- are the Gaussian`s model parameters.
        sigma=0.01,
    ):
        self.centre = centre
        self.intensity = intensity
        self.sigma = sigma

    def profile_from_xvalues(self, xvalues):
        """
        Calculate the intensity of the light profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues : ndarray
            The x coordinates in the original reference frame of the data.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


# %%
"""
We've extended the `Gaussian` class to have a method `profile_from_xvalues`. Given an input set of x coordinates
this computes the intensity of the `Gaussian` at every point. We've already seen how the data contains the xvalues we 
use, which are a 1D NumPy array spanning values 0 to 100.

If we pass these values to an instance of the `Gaussian` class, we can create a line of the gaussian`s values.
"""

# %%
model = af.PriorModel(Gaussian)
model.centre = af.UniformPrior(lower_limit=0.0, upper_limit=np.inf)
model.intensity = af.UniformPrior(lower_limit=0.0, upper_limit=np.inf)
model.sigma = af.UniformPrior(lower_limit=0.0, upper_limit=np.inf)

gaussian = model.instance_from_vector(vector=[60.0, 20.0, 15.0])

model_data = gaussian.profile_from_xvalues(xvalues=xvalues)

plt.plot(xvalues, model_data)
plt.xlabel("x")
plt.ylabel("Intensity")
plt.show()
plt.clf()

# %%
"""
Different values of centre, intensity and sigma change the Gaussian`s apperance - have a go at editing some of the
values above to see this behaviour.
"""

# %%
"""
Okay, so lets recap. We've defined a model which is a 1D `Gaussian` and given a set of parameters for that model
$(x, I, \sigma)$ we can create `model_data` of the `Gaussian`. And, we have some data of a 1D `Gaussian` we want to
fit this model with, so as to determine the values of $(x, I, \sigma)$ from which it was created. So how do we do that?

Simple, we take the image from our data and our model_image of the `Gaussian` and subtract the two to get a
`residual-map`.
"""

# %%
residual_map = data - model_data
plt.plot(xvalues, model_data)
plt.xlabel("x")
plt.ylabel("Residuals")
plt.show()
plt.clf()

# %%
"""
Clearly, this model isn't a good fit to the data - which was to be expected as they looked nothing alike!

Next, we want to quantify how good (or bad) the fit actually was, via some goodness-of-fit measure. This measure
needs to account for noise in the data - after all if we fit a pixel badly simply because it was very noisy we want
our goodness-of-fit to account for that.

To account for noise, we take our residual-map and divide it by the noise-map, to get the `normalized residual-map`.
"""

# %%
normalized_residual_map = residual_map / noise_map
plt.plot(xvalues, model_data)
plt.xlabel("x")
plt.ylabel("Normalized Residuals")
plt.show()
plt.clf()

# %%
"""
We`re getting close to a goodness-of-fit measure, but there is still a problem - we have negative and positive values
in the normalized residual-map. A value of -0.2 represents just as good of a fit as a value of 0.2, so we want them
to both be the same value.

Thus, we next define a `chi-squared-map`, which is the normalized residual-map squared. This makes negative and
positive values both positive and thus defines them on a common overall scale.
"""

# %%
chi_squared_map = (normalized_residual_map) ** 2
plt.plot(xvalues, model_data)
plt.xlabel("x")
plt.ylabel("Chi-Squareds")
plt.show()
plt.clf()

# %%
"""
Great, even when looking at a chi-squared-map its clear that our model gives a rubbish fit to the data.

Finally, we want to reduce all the information in our chi-squared-map into a single goodness-of-fit measure. To do
this we define the `chi-squared`, which is the sum of all values on the chi-squared-map.
"""

# %%
chi_squared = np.sum(chi_squared_map)
print("Chi-squared = ", chi_squared)

# %%
"""
The lower our chi-squared, the fewer residuals in the fit between our model and the data and therefore the better our 
fit!

From the chi-squared we can then define our final goodness-of-fit measure, the `log_likelihood`, which is the
chi-squared value times -0.5.
"""

# %%
log_likelihood = -0.5 * chi_squared
print("Log Likelihood = ", log_likelihood)

# %%
"""
Why is the log likelihood the chi-squared times -0.5? Lets not worry about. This is simply the standard definition of a
log_likelihood in statistics (it relates to the noise-properties of our data-set). For now, just accept that this is 
what a log likelihood is and if we want to fit a model to data our goal is to thus find the combination of model 
parameters that maximizes our log_likelihood.

There is a second quantity that enters the log likelihood, called the `noise-normalization`. This is the log sum of all
noise values squared in our data (given the noise-map is fixed, the noise_normalization retains the same value for
all models that we fit. Nevertheless, it is good practise to include it in the log likelihood).
"""

# %%
noise_normalization = np.sum(np.log(2 * np.pi * noise_map ** 2.0))

# %%
"""
Again, like the definition of a log likelihood, lets not worry about why a noise normalization is defined in this way 
or why its in our goodness-of-fit. Lets just accept for now that this is how it is in statistics.

Thus, we now have the definition of a log likelihood that we'll use hereafter in all **PyAutoFit** tutorials.
"""

# %%
log_likelihood = -0.5 * (chi_squared + noise_normalization)
print("Log Likelihood = ", log_likelihood)

# %%
"""
If you are familiar with model-fitting, you`ll have probably heard of terms like `residuals`, `chi-squared` and
`log_likelihood` before. These are the standard metrics by which a model-fit`s quality is measured. They are used for
model fitting in general, so not just when your data is 1D but when its a 2D image, 3D datacube or something else
entirely!

If you haven`t performed model fitting before and these terms are new to you, make sure you are clear on exactly what
they all mean as they are at the core of all model-fitting performed in **PyAutoFit**!
"""

# %%
"""
So to recap the previous tutorial and this one:
    
 - We can define a model components in **PyAutoFit**, like our `Gaussian`, using Python classes that follow a certain 
 format.
 
 - The model component`s parameters each have priors, which given a unit vector can be mapped to an instance of the
 `Gaussian` class.
 
 - We can use this model instance to create model data of our `Gaussian` and compare it to data and quantify the
 goodness-of-fit via a log likelihood.

Thus we have everything we need to fit our model to our data! So, how do we go about finding the best-fit model?
That is, the model which maximizes the log likelihood.

The most simple thing we can do is guess parameters, and when we guess parameters that give a good fit, guess another
set of parameters near those values. We can then repeat this process, over and over, until we find a really good model!

For our `Gaussian` this works pretty well, below I`ve fitted 5 diferent `Gaussian` models and ended up landing on
the best-fit model (the model I used to create the `Dataset` in the first place!).

For convenience, I`ve create functions which compute the chi-squared-map and log likelihood of a model-fit, alongside a
method to plot a profile, residual-map or chi-squared-map.
"""

# %%
def chi_squared_map_from_data_and_model_data(data, noise_map, model_data):
    residual_map = data - model_data
    normalized_residual_map = residual_map / noise_map
    return (normalized_residual_map) ** 2


def log_likelihood_from_data_and_model_data(data, noise_map, model_data):

    chi_squared_map = chi_squared_map_from_data_and_model_data(
        data=data, noise_map=noise_map, model_data=model_data
    )
    chi_squared = sum(chi_squared_map)
    noise_normalization = np.sum(np.log(2 * np.pi * noise_map ** 2.0))
    log_likelihood = -0.5 * (chi_squared + noise_normalization)

    return log_likelihood


def plot_line(xvalues, line, ylabel=None):

    plt.plot(xvalues, line)
    plt.xlabel("x")
    plt.ylabel(ylabel)
    plt.show()
    plt.clf()


# %%
gaussian = model.instance_from_vector(vector=[50.0, 10.0, 5.0])
model_data = gaussian.profile_from_xvalues(xvalues=xvalues)
chi_squared_map = chi_squared_map_from_data_and_model_data(
    data=data, noise_map=noise_map, model_data=model_data
)
plot_line(xvalues=xvalues, line=chi_squared_map, ylabel="Chi-Squareds")

log_likelihood = log_likelihood_from_data_and_model_data(
    data=data, noise_map=noise_map, model_data=model_data
)
print("Log Likelihood:")
print(log_likelihood)

gaussian = model.instance_from_vector(vector=[50.0, 25.0, 5.0])
model_data = gaussian.profile_from_xvalues(xvalues=xvalues)
chi_squared_map = chi_squared_map_from_data_and_model_data(
    data=data, noise_map=noise_map, model_data=model_data
)
plot_line(xvalues=xvalues, line=chi_squared_map, ylabel="Chi-Squareds")

log_likelihood = log_likelihood_from_data_and_model_data(
    data=data, noise_map=noise_map, model_data=model_data
)
print("Log Likelihood:")
print(log_likelihood)

gaussian = model.instance_from_vector(vector=[50.0, 25.0, 10.0])
model_data = gaussian.profile_from_xvalues(xvalues=xvalues)
chi_squared_map = chi_squared_map_from_data_and_model_data(
    data=data, noise_map=noise_map, model_data=model_data
)
plot_line(xvalues=xvalues, line=chi_squared_map, ylabel="Chi-Squareds")

log_likelihood = log_likelihood_from_data_and_model_data(
    data=data, noise_map=noise_map, model_data=model_data
)
print("Log Likelihood:")
print(log_likelihood)

# %%
"""
You can now perform model-fitting with **PyAutoFit**! All we have to do is guess lots of parameters, over and over and
over again, until we hit a model with a high log_likelihood. Yay!

Of course, you`re probably thinking, is that really it? Should we really be guessing models to find the best-fit?

Obviously, the answer is no. Imagine our model was more complex, that it had many more parameters than just 4.
Our approach of guessing parameters won`t work - it could take days, maybe years, to find models with a high
log_likelihood, and how could you even be sure they ware the best-fit models? Maybe a set of parameters you never tried
provide an even better fit?

Of course, there is a much better way to perform model-fitting, and in the next tutorial we'll take you through how
to do such fitting in **PyAutoFit**, using a `_NonLinearSearch_`.

__Your Model__

To end, its worth quickly thinking about the model you ultimately want to fit with **PyAutoFit**. In this example,
we extended the `Gaussian` class to contain the function we needed to generate an image of the `Gaussian` and thus
generate the model-image we need to fit our data. For your model fitting problem can you do something similar?
Or is your model-fitting task a bit more complicated than this? Maybe there are more model component you want to
combine or there is an inter-dependency between models?

**PyAutoFit** provides a lot of flexibility in how you ultimate use your model instances, so whatever your problem you
should find that it is straight forward to find a solution. But, whatever you need to do at its core your modeling
problem will break down into the tasks we did in this tutorial:

- Use your model to create some model data.
- Subtract it from the data to create residuals.
- Use these residuals in conjunction with your noise-map to define a log likelihood.
- Find the highest log likelihood models.

So, get thinking about how these steps would be performed for your model!
"""
